@@ -13,6 +13,16 @@ import math
 version = csrbox.__version__
 logger = logging.getLogger(__name__)
 
+def priv2bits (priv_mode):
+    if priv_mode == 'M':
+        return 3
+    if priv_mode == 'S':
+        return 1
+    if priv_mode == 'U':
+        return 0
+    logger.error('Unsupported priv mode found')
+    raise SystemExit
+
 def csr_gen (isa_file, grp_file, custom_file, debug_file, custom_attr_file, work_dir,
         probeinst, logging=False):
     '''
@@ -138,11 +148,15 @@ def groups_gen(isa_yaml, grp_yaml, attr_yaml, work_dir, probeinst, logging=False
     xlen = str(max(isa_yaml['hart0']['supported_xlen']))
 
     station_isa = {}
+    aux_info = {}
     csr_isa = {}
 
     not_acc = []
     max_len = len(grp_yaml)
 
+    aux_info['custom_exceptions']=isa_yaml['hart0']['custom_exceptions']
+    aux_info['custom_interrupts']=isa_yaml['hart0']['custom_interrupts']
+
     for grp, elems in grp_yaml.items():
         if elems != None:
           if 'grp'+str(max_len) == grp:
@@ -169,24 +183,24 @@ def groups_gen(isa_yaml, grp_yaml, attr_yaml, work_dir, probeinst, logging=False
         pkg_name = filename[:-4]
         bsvstr += str(license_header.format(version, str(datetime.datetime.now())))
         bsvstr += (package_temp.format(pkg_name))
-        bsvstr = station_interface(station_isa, grp_yaml, attr_yaml, rvxlen,
+        bsvstr = station_interface(station_isa, grp_yaml, attr_yaml, aux_info, rvxlen,
                 pkg_name, last, bsvstr, logging)
         bsvstr = station_module(station_isa, grp_yaml, attr_yaml, rvxlen,
                 pkg_name, last, bsvstr, logging)
-        bsvstr = station_regs(station_isa, grp_yaml, attr_yaml, rvxlen,
+        bsvstr = station_regs(station_isa, grp_yaml, attr_yaml, aux_info, rvxlen,
                 pkg_name, last, bsvstr, logging)
         bsvstr = station_modbody(station_isa, grp_yaml, attr_yaml, rvxlen,
                 pkg_name, last, bsvstr, logging)
         bsvstr = station_req(station_isa, grp_yaml, attr_yaml, rvxlen,
                 pkg_name, last, bsvstr, logging)
-        bsvstr = station_methods(station_isa, grp_yaml, attr_yaml, rvxlen,
+        bsvstr = station_methods(station_isa, grp_yaml, attr_yaml, aux_info, rvxlen,
                 pkg_name, last, bsvstr, logging)
         bsvstr += (end_module_temp)
         bsvstr +=(end_package_temp)
         bsvfile.write(bsvstr)
         station_isa.clear()  
     physical_addr_sz = isa_yaml['hart0']['physical_addr_sz']
-    bsvstr = csr_top(csr_isa, grp_yaml, attr_yaml, rvxlen, 
+    bsvstr = csr_top(csr_isa, grp_yaml, attr_yaml, aux_info, rvxlen, 
              'csrbox', logging)
     filename = 'csrbox.bsv'
     bsvfile = open(work_dir + filename, 'w')
@@ -318,6 +332,10 @@ def csr_define(isa_yaml, work_dir, bsv, logging=False):
         for ce in isa_yaml['hart0']['custom_interrupts']:
             definefile.write('\n`define {0} {1}'.format(ce['cause_name'],
                 ce['cause_val']))
+    if 'parking_loop' in isa_yaml['hart0'] and \
+            isa_yaml['hart0']['parking_loop'] is not None:
+        definefile.write("\n`define debug_parking_loop "+ \
+                str(isa_yaml['hart0']['parking_loop'])+"\n")
     definefile.close()
 
 
@@ -337,15 +355,15 @@ def create_concat_vector(mylist, xlen):
     cv = ' '
     last_lsb = xlen - 1
     while len(newlist) != 0:
-        msb, width, name, type_reg = newlist[0]
+        msb, width, name, type_reg, prefix = newlist[0]
         newlist.pop(0)
         if msb != last_lsb:
             cv += readonly_temp.format(last_lsb - msb) + ','
             concat_length += 1
         if 'ro_variable' in type_reg and (name != 'mstatus_sd'):
-            cv += 'readOnlyReg(rg_' + str(name) + ' ),'
+            cv += 'readOnlyReg({0}{1}), '.format(prefix , str(name))
         else:
-            cv += 'rg_' + str(name) + ' ,'
+            cv += prefix + str(name) + ' ,'
         concat_length += 1
         last_lsb = msb - width
 
@@ -488,7 +506,7 @@ def derive_warl_func(field_name, width, warl, csr_dep):
     bsv_str = bsv_str[:-22]
     return bsv_str
 
-def station_interface(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
+def station_interface(isa_yaml, grp_yaml, attr_yaml, aux_info, rvxlen, pkg_name,
         last, bsvstr, logging = False):
 
     '''
@@ -615,10 +633,18 @@ def station_interface(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
                     bsvstr +=(methoda_read_temp.format('Bit#('+str(size)
                             + ') _' + shadow_csr, shadow_csr))
                     incomming_reads.append(shadow_csr)
+
+    # create methods for custom interrupts
+    if 'mip' in isa_yaml:
+        for ci in aux_info['custom_interrupts']:
+            bsvstr += '\n    method Action ma_set_mip_{0} (Bit#(1) _{0});'.format( \
+                ci['cause_name'])
     if logging:
         logger.info('Interface created for : ' + pkg_name)
     # define generic methods for all stations
     bsvstr +=(method_station_generic)
+    if debug_enabled:
+        bsvstr += method_station_generic_debug
     if not last:
         bsvstr +=(method_fwd_temp)
     bsvstr += (end_interface_temp)
@@ -697,7 +723,7 @@ def station_modbody (isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
         logger.info('Custom rules/module elements created for : ' + pkg_name)
     return bsvstr
 
-def station_methods(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
+def station_methods(isa_yaml, grp_yaml, attr_yaml, aux_info, rvxlen, pkg_name,
         last, bsvstr, logging = False):
     ''' 
         Function to create the bluespec method definitions for each
@@ -723,6 +749,9 @@ def station_methods(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
         :type bsvstr: str
         :type work_dir: str
     '''
+
+    global debug_enabled
+
     xlen = int(rvxlen[2:])
     if logging:
         logger.info('Creating method definitions for : ' + pkg_name)
@@ -828,8 +857,18 @@ def station_methods(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
                         textwrap.indent(add_ifc, '    ') + '\n'
                     req_ifc.append(add_ifc)
 
+    # create methods for custom interrupts
+    if 'mip' in isa_yaml:
+        for ci in aux_info['custom_interrupts']:
+            bsvstr += '''
+    method Action ma_set_mip_{0} (Bit#(1) _{0});
+        rg_mip_{0} <= _{0};
+    endmethod:ma_set_mip_{0}
+'''.format(ci['cause_name'])
 
     bsvstr += method_generic_def_temp
+    if debug_enabled:
+        bsvstr += method_generic_debug_def_temp
     if not last:
         bsvstr += method_fwd_def_temp
     return bsvstr
@@ -931,7 +970,7 @@ def station_req(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
 
     return bsvstr
 
-def station_regs(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
+def station_regs(isa_yaml, grp_yaml, attr_yaml, aux_info, rvxlen, pkg_name,
         last, bsvstr, logging = False):
     '''
         Function to create the register and wires required to create the csrs.
@@ -959,9 +998,13 @@ def station_regs(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
         :type bsvstr: str
         :type work_dir: str
     '''
+    global debug_enabled
+        
     if logging:
         logger.info('Generating storage elements for : ' + pkg_name)
     bsvstr += (other_vars_temp)
+    if debug_enabled:
+        bsvstr += other_vars_debug_temp
     shadows = []
     xlen = int(rvxlen[2:])
     # creating additional required registers
@@ -1056,16 +1099,25 @@ def station_regs(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
                     subdescription = subcontent['description']
                     width = msb - lsb + 1
                     subreset_val = get_reset_val(reset_val, width, lsb, str(xlen))
+                    storage_type = 'rg_'
                     if subcontent['shadow'] is not None:
-                        concat.append([ msb, width, subcontent['shadow'].replace('.','_'), ''])
+                        if subcontent['shadow_type'] == 'ro':
+                            type_reg = 'ro_variable'
+                        else:
+                            type_reg = 'warl'
+                        concat.append([ msb, width,
+                            subcontent['shadow'].replace('.','_'), type_reg,
+                            storage_type])
                     else:
                         concat.append([ msb, width, csrname + '_' + subfield, 
-                         type_reg])
+                         type_reg, storage_type])
 
                     if subfield == 'sd' and csrname == 'mstatus':
                         bsvstr += temp_readOnlyRegister.format(subdescription,
                                 str(width), csrname+'_'+subfield, 
                                 "pack((rg_mstatus_xs == 2\'b11) || (rg_mstatus_fs == 2\'b11))")
+                    elif subfield in ['meip','msip','mtip']:
+                        bsvstr += temp_wire.format(width,csrname+'_'+subfield)
                     elif subcontent['shadow'] is not None:
                         shadow_csr = subcontent['shadow'].split('.')[0]
                         if shadow_csr not in isa_yaml and shadow_csr not in incomming_reads and\
@@ -1097,6 +1149,18 @@ def station_regs(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
                     else:
                         bsvstr += temp_readOnlyRegister.format(subdescription,
                                 str(width), csrname+'_'+subfield, 0)
+            if csrname.lower() in ['mie']:
+                for ci in aux_info['custom_interrupts']:
+                    bsvstr += temp_register.format('',1,csrname+'_'+ci['cause_name'],ci['on_reset_enable'])
+                    concat.append([ci['cause_val'],1,csrname+'_'+ci['cause_name'],'','rg_'])
+            if csrname.lower() in ['sie']:
+                for ci in aux_info['custom_interrupts']:
+                    if ci['priv_mode']== 'S':
+                        concat.append([ci['cause_val'],1,'mie_'+ci['cause_name'],'','rg_'])
+            if csrname.lower() in ['mip']:
+                for ci in aux_info['custom_interrupts']:
+                    bsvstr += temp_wire.format(1,csrname+'_'+ci['cause_name'])
+                    concat.append([ci['cause_val'],1,csrname+'_'+ci['cause_name'],'ro_variable', 'rg_'])
             length, cv = create_concat_vector(concat, int(xlen))
             bsvstr += concat_temp.format(description, xlen, csrname,\
                       length, cv)
@@ -1121,7 +1185,7 @@ def find_group(bar, csrname, logging=False):
             return grps
     return None
 
-def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
+def csr_top(isa_yaml, grp_yaml, attr_yaml, aux_info, rvxlen, pkgname, logging = False):
     '''
         Function to create the top level bluespec module which instantiates the
         individual group/station and creates the connection between them.
@@ -1176,11 +1240,11 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
     sideleg_grp = find_group(grp_yaml, 'sideleg')
     dpc_grp = find_group(grp_yaml,'dpc')
     dcsr_grp_name = find_group(grp_yaml,'dcsr')
-    dtvec_grp = find_group(grp_yaml,'dtvec')
+    mip_grp_name = find_group(grp_yaml, 'mip')
 
     if debug_enabled:
-        if dpc_grp is None or dcsr_grp_name is None or dtvec_grp is None:
-            logger.error('Debug is enabled but one of dpc, dcsr and/or dtvec is not available in any group')
+        if dpc_grp is None or dcsr_grp_name is None:
+            logger.error('Debug is enabled but one/both of dpc, dcsr is not available in any group')
             sys.exit(1)
     
     pmp_entries = 0
@@ -1264,6 +1328,9 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
                     args = arguments
                 bsvstr +=('\n    '+args)
 
+    for ci in aux_info['custom_interrupts']:
+        bsvstr += '\n    method Action ma_set_mip_{0} (Bit#(1) _{0});'.format( \
+                ci['cause_name'])
     # debug specific interfaces
     if debug_enabled:
         bsvstr += method_top_debug_dec;
@@ -1279,10 +1346,7 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
     if 'minstret' in isa_yaml:
         if isa_yaml['minstret']['rv'+str(xlen)]['accessible']:
             mod_attr_str = '(*conflict_free="ma_incr_minstret,sbread.mv_csr_minstret"*)'
-    if debug_enabled:
-        extra_reset = '#(Reset debug_reset)'
-    else:
-        extra_reset = ''
+    extra_reset = ''
     bsvstr += module_temp.format(pkgname,mod_attr_str, extra_reset) + '\n'
 
     if logging:
@@ -1302,12 +1366,8 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
     
     if debug_enabled:
         bsvstr += '''
-    /*doc:reg:The register indicates that the core is halted*/
-    Reg#(Bit#(1)) rg_core_halted <- mkReg(0);
-
-    /*doc:reg: this register indicates if the core has been reset by the
-    debugger*/
-    Reg#(Bit#(1)) rg_core_has_reset <- mkRegA(0, reset_by debug_reset);
+    /*doc:reg:The register indicates that the core is in debug mode*/
+    Reg#(Bit#(1)) rg_debug_mode <- mkReg(0);
 '''
 
     anyhit = ''
@@ -1346,14 +1406,9 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
     # -- connect rg_prv to all stations
     for grp in grp_yaml:
         bsvstr += temp_prv_connection.format(grp)
+        if debug_enabled:
+            bsvstr += temp_dbg_connection.format(grp)
 
-    if debug_enabled:
-        bsvstr += '''
-    rule rl_core_reset;
-        rg_core_has_reset <= 1;
-    endrule: rl_core_reset
-'''
-    
     # create action methods of dependent csrs
     if logging:
         logger.debug('-- defining action methods')
@@ -1408,6 +1463,9 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
         logger.debug('-- Defining trap and return methods')
     bsvstr += method_top_def_temp
     bsvstr += method_top_ret_temp.format(xlen)
+    if debug_enabled:
+        bsvstr += method_top_ret_debug_temp.format(dpc_grp, dcsr_grp_name,
+                mstatus_grp);
     if usertraps :
         bsvstr += method_top_ret_user_temp.format(mstatus_grp, uepc_grp)
     if supervisor:
@@ -1421,8 +1479,7 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
     if (usertraps or supervisor) and medeleg_grp is not None:
         bsvstr += method_top_trap_medeleg.format(medeleg_grp, mideleg_grp)
     if debug_enabled:
-        bsvstr += method_top_trap_debug_temp.format(dpc_grp, dcsr_grp_name,
-                dtvec_grp)
+        bsvstr += method_top_trap_debug_temp.format(dpc_grp, dcsr_grp_name)
     if supervisor:
         bsvstr += method_top_trap_supervisor_temp.format(stval_grp, sepc_grp, scause_grp,
             mstatus_grp, stvec_grp)
@@ -1512,9 +1569,12 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
                 methods = re.findall(r'\s*method\s*Action\s*(.*?)[\s(]', args)
                 for x in methods:
                     bsvstr += methoda_top_temp.format(x, csr_grp)
+    for ci in aux_info['custom_interrupts']:
+        bsvstr += '\n    method ma_set_mip_{0} = {1}.ma_set_mip_{0};'.format( \
+                ci['cause_name'], mip_grp_name)
     # debug specific interfaces
     if debug_enabled:
-        bsvstr += method_top_debug_def;
+        bsvstr += method_top_debug_def.format(dcsr_grp_name);
             
 
     # -- end the module
